import * as path from 'path';
import * as urlParser from 'url';
import deepmerge from 'deepmerge';
import * as backoff from 'backoff';
import { default as imagemin } from 'imagemin';
import imageminAdvPng from 'imagemin-advpng';
import axios from 'axios';
import { default as imageminPngquant } from 'imagemin-pngquant';
import imageminGifsicle from 'imagemin-gifsicle';
import imageminJpegoptim from 'imagemin-jpegoptim';
import imageminWebp from 'imagemin-webp';
import sharp from 'sharp';
import http from 'http';
import https from 'https';
import { normalizeMwResponse, objToQueryString, DB_ERROR, WEAK_ETAG_REGEX, renderArticle, stripHttpFromUrl, isBitmapImageMimeType, isImageUrl, getMimeType, isWebpCandidateImageMimeType, } from './util/index.js';
import logger from './Logger.js';
const imageminOptions = new Map();
imageminOptions.set('default', new Map());
imageminOptions.set('webp', new Map());
imageminOptions.get('default').set('image/png', {
    plugins: [imageminPngquant({ speed: 3, strip: true, dithering: 0 }), imageminAdvPng({ optimizationLevel: 4, iterations: 5 })],
});
imageminOptions.get('default').set('image/jpeg', {
    plugins: [imageminJpegoptim({ max: 60, stripAll: true })],
});
imageminOptions.get('default').set('image/gif', {
    plugins: [imageminGifsicle({ optimizationLevel: 3, colors: 64 })],
});
imageminOptions.get('webp').set('image/png', {
    plugins: [imageminWebp({ quality: 50, method: 6 })],
});
imageminOptions.get('webp').set('image/jpeg', {
    plugins: [imageminWebp({ quality: 50, method: 6 })],
});
class Downloader {
    mw;
    loginCookie = '';
    speed;
    baseUrl;
    baseUrlForMainPage;
    cssDependenceUrls = {};
    webp = false;
    requestTimeout;
    uaString;
    activeRequests = 0;
    maxActiveRequests = 1;
    urlPartCache = {};
    backoffOptions;
    optimisationCacheUrl;
    s3;
    mwCapabilities; // todo move to MW
    arrayBufferRequestOptions;
    jsonRequestOptions;
    streamRequestOptions;
    constructor({ mw, uaString, speed, reqTimeout, optimisationCacheUrl, s3, webp, backoffOptions }) {
        this.mw = mw;
        this.uaString = uaString;
        this.speed = speed;
        this.maxActiveRequests = speed * 10;
        this.requestTimeout = reqTimeout;
        this.loginCookie = '';
        this.optimisationCacheUrl = optimisationCacheUrl;
        this.webp = webp;
        this.s3 = s3;
        this.mwCapabilities = {
            apiAvailable: false,
            veApiAvailable: false,
            coordinatesAvailable: true,
            desktopRestApiAvailable: false,
            mobileRestApiAvailable: false,
        };
        this.backoffOptions = {
            strategy: new backoff.ExponentialStrategy(),
            failAfter: 7,
            retryIf: (err) => err.code === 'ECONNABORTED' || ![400, 403, 404].includes(err.response?.status),
            backoffHandler: (number, delay) => {
                logger.info(`[backoff] #${number} after ${delay} ms`);
            },
            ...backoffOptions,
        };
        this.arrayBufferRequestOptions = {
            // HTTP agent pools with 'keepAlive' to reuse TCP connections, so it's faster
            httpAgent: new http.Agent({ keepAlive: true }),
            httpsAgent: new https.Agent({ keepAlive: true }),
            headers: {
                'cache-control': 'public, max-stale=86400',
                'user-agent': this.uaString,
                cookie: this.loginCookie,
            },
            responseType: 'arraybuffer',
            timeout: this.requestTimeout,
            method: 'GET',
            validateStatus(status) {
                return (status >= 200 && status < 300) || status === 304;
            },
        };
        this.jsonRequestOptions = {
            // HTTP agent pools with 'keepAlive' to reuse TCP connections, so it's faster
            httpAgent: new http.Agent({ keepAlive: true }),
            httpsAgent: new https.Agent({ keepAlive: true }),
            headers: {
                accept: 'application/json',
                'cache-control': 'public, max-stale=86400',
                'accept-encoding': 'gzip, deflate',
                'user-agent': this.uaString,
                cookie: this.loginCookie,
            },
            responseType: 'json',
            timeout: this.requestTimeout,
            method: 'GET',
        };
        this.streamRequestOptions = {
            // HTTP agent pools with 'keepAlive' to reuse TCP connections, so it's faster
            httpAgent: new http.Agent({ keepAlive: true }),
            httpsAgent: new https.Agent({ keepAlive: true }),
            headers: {
                accept: 'application/octet-stream',
                'cache-control': 'public, max-stale=86400',
                'accept-encoding': 'gzip, deflate',
                'user-agent': this.uaString,
                cookie: this.loginCookie,
            },
            responseType: 'stream',
            timeout: this.requestTimeout,
            method: 'GET',
        };
    }
    serializeUrl(url) {
        const { path } = urlParser.parse(url);
        const cacheablePart = url.replace(path, '');
        const cacheEntry = Object.entries(this.urlPartCache).find(([, value]) => value === cacheablePart);
        let cacheKey;
        if (!cacheEntry) {
            const cacheId = String(Object.keys(this.urlPartCache).length + 1);
            this.urlPartCache[cacheId] = cacheablePart;
            cacheKey = `_${cacheId}_`;
        }
        else {
            cacheKey = `_${cacheEntry[0]}_`;
        }
        return `${cacheKey}${path}`;
    }
    deserializeUrl(url) {
        if (!url.startsWith('_'))
            return url;
        const [, cacheId, ...pathParts] = url.split('_');
        const path = pathParts.join('_');
        const cachedPart = this.urlPartCache[cacheId];
        return `${cachedPart}${path}`;
    }
    async setBaseUrls() {
        this.baseUrl = this.mwCapabilities.mobileRestApiAvailable
            ? this.mw.mobileRestApiUrl.href
            : this.mwCapabilities.desktopRestApiAvailable
                ? this.mw.desktopRestApiUrl.href
                : this.mwCapabilities.veApiAvailable
                    ? this.mw.veApiUrl.href
                    : undefined;
        this.baseUrlForMainPage = this.mwCapabilities.desktopRestApiAvailable ? this.mw.desktopRestApiUrl.href : this.mwCapabilities.veApiAvailable ? this.mw.veApiUrl.href : undefined;
        logger.log('Base Url: ', this.baseUrl);
        logger.log('Base Url for Main Page: ', this.baseUrlForMainPage);
        if (!this.baseUrl || !this.baseUrlForMainPage)
            throw new Error('Unable to find appropriate API end-point to retrieve article HTML');
    }
    async checkApiAvailabilty(url) {
        try {
            const resp = await axios.get(url, { headers: { cookie: this.loginCookie } });
            // Check for hostname is for domain name in cases of redirects.
            return resp.status === 200 && !resp.headers['mediawiki-api-error'] && path.dirname(url) === path.dirname(resp.request.res.responseUrl);
        }
        catch (err) {
            return false;
        }
    }
    async checkCapabilities(testArticleId = 'MediaWiki:Sidebar') {
        // By default check all API's responses and set the capabilities
        // accordingly. We need to set a default page (always there because
        // installed per default) to request the REST API, otherwise it would
        // fail the check.
        this.mwCapabilities.mobileRestApiAvailable = await this.checkApiAvailabilty(this.mw.getMobileRestApiArticleUrl(testArticleId));
        this.mwCapabilities.desktopRestApiAvailable = await this.checkApiAvailabilty(this.mw.getDesktopRestApiArticleUrl(testArticleId));
        this.mwCapabilities.veApiAvailable = await this.checkApiAvailabilty(this.mw.getVeApiArticleUrl(testArticleId));
        this.mwCapabilities.apiAvailable = await this.checkApiAvailabilty(this.mw.apiUrl.href);
        // Coordinate fetching
        const reqOpts = objToQueryString({
            ...this.getArticleQueryOpts(),
        });
        const resp = await this.getJSON(`${this.mw.apiUrl.href}${reqOpts}`);
        const isCoordinateWarning = resp.warnings && resp.warnings.query && (resp.warnings.query['*'] || '').includes('coordinates');
        if (isCoordinateWarning) {
            logger.info('Coordinates not available on this wiki');
            this.mwCapabilities.coordinatesAvailable = false;
        }
    }
    removeEtagWeakPrefix(etag) {
        return etag && etag.replace(WEAK_ETAG_REGEX, '');
    }
    query(query) {
        return this.getJSON(this.mw.getApiQueryUrl(query));
    }
    async getArticleDetailsIds(articleIds, shouldGetThumbnail = false) {
        let continuation;
        let finalProcessedResp;
        while (true) {
            const queryOpts = {
                ...this.getArticleQueryOpts(shouldGetThumbnail, true),
                titles: articleIds.join('|'),
                ...(this.mwCapabilities.coordinatesAvailable ? { colimit: 'max' } : {}),
                ...(this.mw.getCategories
                    ? {
                        cllimit: 'max',
                        clshow: '!hidden',
                    }
                    : {}),
                ...(continuation || {}),
            };
            const queryString = objToQueryString(queryOpts);
            const reqUrl = this.mw.getApiQueryUrl(queryString);
            const resp = await this.getJSON(reqUrl);
            Downloader.handleMWWarningsAndErrors(resp);
            let processedResponse = resp.query ? normalizeMwResponse(resp.query) : {};
            if (resp.continue) {
                continuation = resp.continue;
                const relevantDetails = this.stripNonContinuedProps(processedResponse);
                finalProcessedResp = finalProcessedResp === undefined ? relevantDetails : deepmerge(finalProcessedResp, relevantDetails);
            }
            else {
                if (this.mw.getCategories) {
                    processedResponse = await this.setArticleSubCategories(processedResponse);
                }
                finalProcessedResp = finalProcessedResp === undefined ? processedResponse : deepmerge(finalProcessedResp, processedResponse);
                break;
            }
        }
        return finalProcessedResp;
    }
    async getArticleDetailsNS(ns, gapcontinue = '') {
        let queryContinuation;
        let finalProcessedResp;
        let gCont = null;
        while (true) {
            const queryOpts = {
                ...this.getArticleQueryOpts(),
                ...(this.mwCapabilities.coordinatesAvailable ? { colimit: 'max' } : {}),
                ...(this.mw.getCategories
                    ? {
                        cllimit: 'max',
                        clshow: '!hidden',
                    }
                    : {}),
                rawcontinue: 'true',
                generator: 'allpages',
                gapfilterredir: 'nonredirects',
                gaplimit: 'max',
                gapnamespace: String(ns),
                gapcontinue,
            };
            if (queryContinuation) {
                queryOpts.cocontinue = queryContinuation?.coordinates?.cocontinue ?? queryOpts.cocontinue;
                queryOpts.clcontinue = queryContinuation?.categories?.clcontinue ?? queryOpts.clcontinue;
                queryOpts.picontinue = queryContinuation?.pageimages?.picontinue ?? queryOpts.picontinue;
                queryOpts.rdcontinue = queryContinuation?.redirects?.rdcontinue ?? queryOpts.rdcontinue;
            }
            const queryString = objToQueryString(queryOpts);
            const reqUrl = this.mw.getApiQueryUrl(queryString);
            const resp = await this.getJSON(reqUrl);
            Downloader.handleMWWarningsAndErrors(resp);
            let processedResponse = normalizeMwResponse(resp.query);
            gCont = resp['query-continue']?.allpages?.gapcontinue ?? gCont;
            const queryComplete = Object.keys(resp['query-continue'] || {}).filter((key) => key !== 'allpages').length === 0;
            if (!queryComplete) {
                queryContinuation = resp['query-continue'];
                const relevantDetails = this.stripNonContinuedProps(processedResponse);
                finalProcessedResp = finalProcessedResp === undefined ? relevantDetails : deepmerge(finalProcessedResp, relevantDetails);
            }
            else {
                if (this.mw.getCategories) {
                    processedResponse = await this.setArticleSubCategories(processedResponse);
                }
                finalProcessedResp = finalProcessedResp === undefined ? processedResponse : deepmerge(finalProcessedResp, processedResponse);
                break;
            }
        }
        return {
            articleDetails: finalProcessedResp,
            gapContinue: gCont,
        };
    }
    async getArticle(articleId, dump, articleDetailXId, articleDetail) {
        const isMainPage = dump.isMainPage(articleId);
        const articleApiUrl = this.getArticleUrl(articleId, isMainPage);
        logger.info(`Getting article [${articleId}] from ${articleApiUrl}`);
        const json = await this.getJSON(articleApiUrl);
        if (json.error) {
            throw json.error;
        }
        return renderArticle(json, articleId, dump, articleDetailXId, this.mwCapabilities, articleDetail);
    }
    async getJSON(_url) {
        const url = this.deserializeUrl(_url);
        await this.claimRequest();
        return new Promise((resolve, reject) => {
            this.backoffCall(this.getJSONCb, url, (err, val) => {
                this.releaseRequest();
                if (err) {
                    const httpStatus = err.response && err.response.status;
                    logger.warn(`Failed to get [${url}] [status=${httpStatus}]`);
                    reject(err);
                }
                else {
                    resolve(val);
                }
            });
        });
    }
    async downloadContent(_url, retry = true) {
        if (!_url) {
            throw new Error(`Parameter [${_url}] is not a valid url`);
        }
        const url = this.deserializeUrl(_url);
        await this.claimRequest();
        try {
            return await new Promise((resolve, reject) => {
                const cb = (err, val) => {
                    if (err) {
                        reject(err);
                    }
                    else {
                        resolve(val);
                    }
                };
                if (retry) {
                    this.backoffCall(this.getContentCb, url, cb);
                }
                else {
                    this.getContentCb(url, cb);
                }
            });
        }
        catch (err) {
            const httpStatus = err.response && err.response.status;
            logger.warn(`Failed to get [${url}] [status=${httpStatus}]`);
            throw err;
        }
        finally {
            this.releaseRequest();
        }
    }
    async canGetUrl(url) {
        try {
            await axios.get(url);
            return true;
        }
        catch (err) {
            return false;
        }
    }
    getArticleUrl(articleId, isMainPage) {
        return `${isMainPage ? this.baseUrlForMainPage : this.baseUrl}${encodeURIComponent(articleId)}`;
    }
    stripNonContinuedProps(articleDetails, cont = {}) {
        const propsMap = {
            pageimages: ['thumbnail', 'pageimage'],
            coordinates: ['coordinates'],
            categories: ['categories'],
        };
        const keysToKeep = ['subCategories', 'revisions', 'redirects'].concat(Object.keys(cont).reduce((acc, key) => acc.concat(propsMap[key] || []), []));
        const items = Object.entries(articleDetails).map(([aId, detail]) => {
            const newDetail = keysToKeep.reduce((acc, key) => {
                const val = detail[key];
                if (!val) {
                    return acc;
                }
                else {
                    return {
                        ...acc,
                        [key]: val,
                    };
                }
            }, {});
            return [aId, newDetail];
        });
        return items.reduce((acc, [key, detail]) => {
            return { ...acc, [key]: detail };
        }, {});
    }
    static handleMWWarningsAndErrors(resp) {
        if (resp.warnings)
            logger.warn(`Got warning from MW Query ${JSON.stringify(resp.warnings, null, '\t')}`);
        if (resp.error?.code === DB_ERROR)
            throw new Error(`Got error from MW Query ${JSON.stringify(resp.error, null, '\t')}`);
        if (resp.error)
            logger.log(`Got error from MW Query ${JSON.stringify(resp.warnings, null, '\t')}`);
    }
    getArticleQueryOpts(includePageimages = false, redirects = false) {
        const validNamespaceIds = this.mw.namespacesToMirror.map((ns) => this.mw.namespaces[ns].num);
        return {
            action: 'query',
            format: 'json',
            prop: `redirects|revisions${includePageimages ? '|pageimages' : ''}${this.mwCapabilities.coordinatesAvailable ? '|coordinates' : ''}${this.mw.getCategories ? '|categories' : ''}`,
            rdlimit: 'max',
            rdnamespace: validNamespaceIds.join('|'),
            redirects: redirects ? true : undefined,
        };
    }
    async setArticleSubCategories(articleDetails) {
        logger.info('Getting subCategories');
        for (const [articleId, articleDetail] of Object.entries(articleDetails)) {
            const isCategoryArticle = articleDetail.ns === 14;
            if (isCategoryArticle) {
                const categoryMembers = await this.getSubCategories(articleId);
                articleDetails[articleId].subCategories = categoryMembers.slice();
            }
        }
        return articleDetails;
    }
    async claimRequest() {
        if (this.activeRequests < this.maxActiveRequests) {
            this.activeRequests += 1;
            return null;
        }
        else {
            await new Promise((resolve) => {
                setTimeout(resolve, 200);
            });
            return this.claimRequest();
        }
    }
    async releaseRequest() {
        this.activeRequests -= 1;
        return null;
    }
    getJSONCb = (url, handler) => {
        logger.info(`Getting JSON from [${url}]`);
        axios
            .get(url, this.jsonRequestOptions)
            .then((a) => handler(null, a.data), handler)
            .catch((err) => {
            try {
                if (err.response && err.response.status === 429) {
                    logger.log('Received a [status=429], slowing down');
                    const newMaxActiveRequests = Math.max(this.maxActiveRequests - 1, 1);
                    logger.log(`Setting maxActiveRequests from [${this.maxActiveRequests}] to [${newMaxActiveRequests}]`);
                    this.maxActiveRequests = newMaxActiveRequests;
                    return this.getJSONCb(url, handler);
                }
                else if (err.response && err.response.status === 404) {
                    handler(err);
                }
            }
            catch (a) {
                logger.log('ERR', err);
                handler(err);
            }
        });
    };
    async getCompressedBody(resp) {
        if (isBitmapImageMimeType(resp.headers['content-type'])) {
            if (isWebpCandidateImageMimeType(this.webp, resp.headers['content-type']) && !this.cssDependenceUrls.hasOwnProperty(resp.config.url)) {
                resp.data = await imagemin
                    .buffer(resp.data, imageminOptions.get('webp').get(resp.headers['content-type']))
                    .catch(async (err) => {
                    if (/Unsupported color conversion request/.test(err.stderr)) {
                        return await imagemin
                            .buffer(await sharp(resp.data).toColorspace('srgb').toBuffer(), imageminOptions.get('webp').get(resp.headers['content-type']))
                            .catch(() => {
                            return resp.data;
                        })
                            .then((data) => {
                            resp.headers['content-type'] = 'image/webp';
                            return data;
                        });
                    }
                    else {
                        return await imagemin.buffer(resp.data, imageminOptions.get('default').get(resp.headers['content-type'])).catch(() => {
                            return resp.data;
                        });
                    }
                })
                    .then((data) => {
                    resp.headers['content-type'] = 'image/webp';
                    return data;
                });
                resp.headers.path_postfix = '.webp';
            }
            else {
                resp.data = await imagemin.buffer(resp.data, imageminOptions.get('default').get(resp.headers['content-type'])).catch(() => {
                    return resp.data;
                });
            }
            return true;
        }
        return false;
    }
    getContentCb = async (url, handler) => {
        logger.info(`Downloading [${url}]`);
        try {
            if (this.optimisationCacheUrl && isImageUrl(url)) {
                this.downloadImage(url, handler);
            }
            else {
                const resp = await axios(url, this.arrayBufferRequestOptions);
                await this.getCompressedBody(resp);
                handler(null, {
                    responseHeaders: resp.headers,
                    content: resp.data,
                });
            }
        }
        catch (err) {
            try {
                this.errHandler(err, url, handler);
            }
            catch (a) {
                handler(err);
            }
        }
    };
    async downloadImage(url, handler) {
        try {
            this.s3
                .downloadBlob(stripHttpFromUrl(url), this.webp ? 'webp' : '1')
                .then(async (s3Resp) => {
                if (s3Resp?.Metadata?.etag) {
                    this.arrayBufferRequestOptions.headers['If-None-Match'] = this.removeEtagWeakPrefix(s3Resp.Metadata.etag);
                }
                const mwResp = await axios(url, this.arrayBufferRequestOptions);
                /* TODO: Code to remove in a few months (February 2023). For
                some reason, it seems a few pictures have 'image/webp'
                mime-type in S3 although they are png, ... This leads to
                having the following code not assuming they should be
                converted to wepb. To avoid this, as a temporary solution,
                such scneario are ignored and mime-type definition relies
                only on url and Mediawiki header. */
                if (s3Resp?.Metadata?.contenttype === 'image/webp') {
                    s3Resp.Metadata.contenttype = undefined;
                }
                // sanitize Content-Type
                mwResp.headers['content-type'] = getMimeType(url, s3Resp?.Metadata?.contenttype || mwResp.headers['content-type']);
                // Most of the images after uploading once will always have
                // 304 status, until modified.
                // 304 does not have to answer with content-type, we have to get it
                // via S3 metadata or extension
                if (mwResp.status === 304) {
                    // eslint-disable-next-line @typescript-eslint/no-unused-vars
                    const headers = (({ Body, ...o }) => o)(s3Resp);
                    if (isWebpCandidateImageMimeType(this.webp, mwResp.headers['content-type']) && !this.cssDependenceUrls.hasOwnProperty(mwResp.config.url)) {
                        headers.path_postfix = '.webp';
                        headers['content-type'] = 'image/webp';
                    }
                    handler(null, {
                        responseHeaders: headers,
                        content: s3Resp.Body,
                    });
                    return;
                }
                // Compress content
                await this.getCompressedBody(mwResp);
                // Check for the etag and upload
                const etag = this.removeEtagWeakPrefix(mwResp.headers.etag);
                if (etag) {
                    this.s3.uploadBlob(stripHttpFromUrl(url), mwResp.data, etag, mwResp.headers['content-type'], this.webp ? 'webp' : '1');
                }
                handler(null, {
                    responseHeaders: mwResp.headers,
                    content: mwResp.data,
                });
            })
                .catch((err) => {
                this.errHandler(err, url, handler);
            });
        }
        catch (err) {
            this.errHandler(err, url, handler);
        }
    }
    errHandler(err, url, handler) {
        if (err.response && err.response.status === 429) {
            logger.log('Received a [status=429], slowing down');
            const newMaxActiveRequests = Math.max(this.maxActiveRequests - 1, 1);
            logger.log(`Setting maxActiveRequests from [${this.maxActiveRequests}] to [${newMaxActiveRequests}]`);
            this.maxActiveRequests = newMaxActiveRequests;
        }
        logger.log(`Not able to download content for ${url} due to ${err}`);
        handler(err);
    }
    async getSubCategories(articleId, continueStr = '') {
        const { query, continue: cont } = await this.getJSON(this.mw.subCategoriesApiUrl(articleId, continueStr));
        const items = query.categorymembers.filter((a) => a && a.title);
        if (cont && cont.cmcontinue) {
            const nextItems = await this.getSubCategories(articleId, cont.cmcontinue);
            return items.concat(nextItems);
        }
        else {
            return items;
        }
    }
    backoffCall(handler, url, callback) {
        const call = backoff.call(handler, url, callback);
        call.setStrategy(this.backoffOptions.strategy);
        call.retryIf(this.backoffOptions.retryIf);
        call.failAfter(this.backoffOptions.failAfter);
        call.on('backoff', this.backoffOptions.backoffHandler);
        call.start();
    }
}
export default Downloader;
//# sourceMappingURL=Downloader.js.map