import { createClient } from 'redis';
import RedisKvs from './util/RedisKvs.js';
import logger from './Logger.js';
class RedisStore {
    _client;
    storesReady;
    _filesToDownloadXPath;
    _filesToRetryXPath;
    _articleDetailXId;
    _redirectsXId;
    constructor(redisPath, opts) {
        const options = { ...opts };
        const quitOnError = !(options.quitOnError === false);
        delete options.quitOnError;
        if (redisPath.startsWith('/') || redisPath.startsWith('./')) {
            options.socket = {
                ...options.socket,
                path: redisPath,
            };
        }
        else {
            options.url = redisPath;
        }
        this._client = createClient(options);
        this._client.on('error', (err) => {
            if (quitOnError) {
                logger.error('Redis Client Error', err);
                process.exit(3);
            }
        });
    }
    async connect(populateStores = true) {
        if (this._client.isOpen) {
            return;
        }
        await this._client.connect();
        if (populateStores) {
            await this.checkForExistingStores();
            await this.populateStores();
            this.storesReady = true;
        }
    }
    async close() {
        if (this._client.isReady && this.storesReady) {
            logger.log('Flushing Redis DBs');
            await Promise.all([this._filesToDownloadXPath.flush(), this._filesToRetryXPath.flush(), this._articleDetailXId.flush(), this._redirectsXId.flush()]);
        }
        if (this._client.isOpen) {
            await this._client.quit();
        }
    }
    async checkForExistingStores() {
        const patterns = ['*-media', '*-media-retry', '*-detail', '*-redirect'];
        let keys = [];
        for (const pattern of patterns) {
            keys = keys.concat(await this._client.keys(pattern));
        }
        keys.forEach(async (key) => {
            try {
                const length = await this._client.hLen(key);
                const time = new Date(Number(key.slice(0, key.indexOf('-'))));
                logger.error(`Found store from previous run from ${time} that is still in redis: ${key} with length ${length}`);
            }
            catch {
                logger.error(`Key ${key} exists in DB, and is no hash.`);
            }
        });
    }
    async populateStores() {
        this._filesToDownloadXPath = new RedisKvs(this._client, `${Date.now()}-media`, {
            u: 'url',
            n: 'namespace',
            m: 'mult',
            w: 'width',
        });
        this._filesToRetryXPath = new RedisKvs(this._client, `${Date.now()}-media-retry`, {
            u: 'url',
            n: 'namespace',
            m: 'mult',
            w: 'width',
        });
        this._articleDetailXId = new RedisKvs(this._client, `${Date.now()}-detail`, {
            s: 'subCategories',
            c: 'categories',
            p: 'pages',
            h: 'thumbnail',
            g: 'coordinates',
            t: 'timestamp',
            r: 'revisionId',
            i: 'internalThumbnailUrl',
            m: 'missing',
            n: 'title',
        });
        this._redirectsXId = new RedisKvs(this._client, `${Date.now()}-redirect`, {
            t: 'targetId',
            n: 'title',
        });
    }
    createRedisKvs(...args) {
        return new RedisKvs(this._client, ...args);
    }
    get client() {
        return this._client;
    }
    get filesToDownloadXPath() {
        return this._filesToDownloadXPath;
    }
    get filesToRetryXPath() {
        return this._filesToRetryXPath;
    }
    get articleDetailXId() {
        return this._articleDetailXId;
    }
    get redirectsXId() {
        return this._redirectsXId;
    }
}
export default RedisStore;
//# sourceMappingURL=RedisStore.js.map