import S3File from 'aws-sdk';
import logger from './Logger.js';
import { Readable } from 'stream';
import { publicIpv4 } from 'public-ip';
class S3 {
    url;
    params;
    s3Handler;
    bucketName;
    constructor(s3Url, s3Params) {
        this.url = s3Url;
        this.params = s3Params;
        this.bucketName = s3Params.bucketName;
    }
    async initialise() {
        const s3UrlBase = new S3File.Endpoint(this.url);
        this.s3Handler = new S3File.S3({
            endpoint: s3UrlBase,
            accessKeyId: this.params.keyId,
            secretAccessKey: this.params.secretAccessKey,
            s3ForcePathStyle: s3UrlBase.protocol === 'http:',
        });
        try {
            if ((await this.bucketExists(this.bucketName)) === true) {
                return true;
            }
        }
        catch (err) {
            throw new Error(`Unable to connect to S3, either S3 login credentials are wrong or bucket cannot be found
                            Bucket used: ${this.bucketName}
                            End point used: ${s3UrlBase.href}
                            Public IP used: ${await publicIpv4()}`);
        }
    }
    async bucketExists(bucket) {
        return new Promise((resolve, reject) => {
            this.s3Handler.headBucket({ Bucket: bucket }, function (err) {
                return err ? reject(err) : resolve(true);
            });
        });
    }
    async uploadBlob(key, data, eTag, contentType, version) {
        const params = {
            Bucket: this.bucketName,
            Key: key,
            Metadata: { etag: eTag, contenttype: contentType, version },
            Body: this.bufferToStream(data),
        };
        try {
            this.s3Handler.upload(params, function (err) {
                if (err) {
                    logger.log(`Not able to upload ${key}: ${err}`);
                }
            });
        }
        catch (err) {
            logger.log('S3 error', err);
        }
    }
    async downloadBlob(key, version = '1') {
        return new Promise((resolve, reject) => {
            this.s3Handler.getObject({ Bucket: this.bucketName, Key: key }, async (err, val) => {
                if (val) {
                    if (val.Metadata.version !== version) {
                        val.Metadata.etag = undefined;
                    }
                    resolve(val);
                }
                else if (err && err.statusCode === 404) {
                    resolve(null);
                }
                else {
                    reject(err);
                }
            });
        }).catch((err) => {
            return err;
        });
    }
    // Only for testing purpose
    async deleteBlob(key) {
        return new Promise((resolve, reject) => {
            this.s3Handler.deleteObject(key, (err, val) => {
                return err ? reject(err) : resolve(val);
            });
        });
    }
    bufferToStream(binary) {
        return new Readable({
            read() {
                this.push(binary);
                this.push(null);
            },
        });
    }
}
export default S3;
//# sourceMappingURL=S3.js.map